// %flair:license{
// This file is part of the Flair framework distributed under the
// CECILL-C License, Version 1.0.
// %flair:license}
//  created:    2014/07/17
//  filename:   V4LCamera.cpp
//
//  author:     Guillaume Sanahuja
//              Copyright Heudiasyc UMR UTC/CNRS 7253
//
//  version:    $Id: $
//
//  purpose:    base class for V4l camera
//
//
/*********************************************************************/

#include "V4LCamera.h"
#include <GroupBox.h>
#include <DoubleSpinBox.h>
#include <CheckBox.h>
#include <Label.h>
#include <Image.h>
#include <FrameworkManager.h>
#include <fcntl.h>
#include <linux/videodev2.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <cstring>
#include <sys/mman.h>
#include <VisionFilter.h>

#define DEFAULT_V4L_BUFFERS 4

#define CLEAR(x) memset (&(x), 0, sizeof (x))

using std::string;
using namespace flair::core;
using namespace flair::gui;

namespace flair {
namespace sensor {
  
V4LCamera::V4LCamera(string name,uint8_t camera_index, uint16_t width, uint16_t height,
                     Image::Type::Format format, uint8_t priority)
    : Thread(getFrameworkManager(), name, priority),
      Camera(name, width, height, format) {
  
  string deviceName="/dev/video"+std::to_string(camera_index);
  device = open(deviceName.c_str(), O_RDWR | O_NONBLOCK);
  if (device == -1) {
      Thread::Err("Cannot open %s\n",deviceName.c_str());
  } else {
      Printf("V4LCamera %s, opened %s\n",name.c_str(),deviceName.c_str());
  }

  /* w/o memset some parts  arent initialized - AKA: Fill it with zeros so it is clean */
  memset(&capture,0,sizeof(CvCaptureCAM_V4L));
  n_buffers = 0;

  
  if(format == Image::Type::Format::UYVY) {
    if(init(width,height,V4L2_PIX_FMT_UYVY) == -1) {
     Thread::Err("initialisation failed\n");
    }
  } else if (format == Image::Type::Format::YUYV) {
    if(init(width,height,V4L2_PIX_FMT_YUYV) == -1) {
     Thread::Err("initialisation failed\n");
    }
  } else {
    Thread::Err("format not supported\n");
  }
  
  /* This is just a technicality, but all buffers must be filled up before any
  staggered SYNC is applied.  SO, filler up. (see V4L HowTo) */

  allocBuffers();

  for (capture.bufferIndex = 0;capture.bufferIndex < ((int)capture.req.count);++capture.bufferIndex) {
    struct v4l2_buffer buf;

    CLEAR (buf);

    buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory      = V4L2_MEMORY_MMAP;
    buf.index       = (unsigned long)capture.bufferIndex;

    if (-1 == xioctl (device, VIDIOC_QBUF, &buf)) {
      Thread::Err("VIDIOC_QBUF error\n");
    }
  }

  /* enable the streaming */
  capture.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  if (-1 == xioctl (device, VIDIOC_STREAMON,&capture.type)) {
    /* error enabling the stream */
    Thread::Err("VIDIOC_STREAMON error\n");
  }
  
  // ground station
  gain = new DoubleSpinBox(GetGroupBox()->NewRow(), "gain:", 0, 1, 0.1);
  exposure = new DoubleSpinBox(GetGroupBox()->LastRowLastCol(), "exposure:", 0,1, 0.1);
  bright = new DoubleSpinBox(GetGroupBox()->LastRowLastCol(), "bright:", 0, 1, 0.1);
  contrast = new DoubleSpinBox(GetGroupBox()->LastRowLastCol(), "contrast:", 0,1, 0.1);
  hue = new DoubleSpinBox(GetGroupBox()->LastRowLastCol(), "hue:", 0, 1, 0.1);
  sharpness = new DoubleSpinBox(GetGroupBox()->LastRowLastCol(), "sharpness:", 0, 1, 0.1);
  sat = new DoubleSpinBox(GetGroupBox()->LastRowLastCol(), "saturation:", 0, 1,0.1);
  autogain = new CheckBox(GetGroupBox()->NewRow(), "autogain:");
  autoexposure = new CheckBox(GetGroupBox()->LastRowLastCol(), "autoexposure:");
  awb = new CheckBox(GetGroupBox()->LastRowLastCol(), "awb:");
  fps = new Label(GetGroupBox()->NewRow(), "fps");

  hasProblems=false;
}

V4LCamera::~V4LCamera() {
  for (int n_buffers = 0; n_buffers < nbBuffers; n_buffers++) {
       //FreeFunction((char*)buffers[n_buffers].start);
  }
  SafeStop();
  Join();
}

int V4LCamera::init(int width, int height,unsigned long colorspace) {
  CLEAR (capture.cap);
  if(-1 == xioctl(device, VIDIOC_QUERYCAP, &capture.cap)) {
    return -1;
  } else {
    CLEAR (capture.capability);
    capture.capability.type = capture.cap.capabilities;

    /* Query channels number */
    if (-1 == xioctl(device, VIDIOC_G_INPUT, &capture.capability.channels)) {
      return -1;
    }
  }
     
  if ((capture.cap.capabilities & V4L2_CAP_VIDEO_CAPTURE) == 0) {
    Thread::Err("device is unable to capture video memory.\n");
    return -1;
  }

  /* Find Window info */
  CLEAR (capture.form);
  capture.form.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      
  if (-1 == xioctl (device, VIDIOC_G_FMT, &capture.form)) {
     Thread::Err("Could not obtain specifics of capture window.\n");
     return -1;
  }
  
  setVideoSize(width, height,colorspace);

  unsigned int min;

  /* Buggy driver paranoia. */
  min = capture.form.fmt.pix.width * 2;
  if (capture.form.fmt.pix.bytesperline < min) {
   capture.form.fmt.pix.bytesperline = min;
  }

  min = capture.form.fmt.pix.bytesperline * capture.form.fmt.pix.height;
  if (capture.form.fmt.pix.sizeimage < min) {
   capture.form.fmt.pix.sizeimage = min;
  }

  return 1;
}

int V4LCamera::setVideoSize(int width, int height,unsigned long colorspace) {

    CLEAR (capture.crop);
    capture.crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    capture.crop.c.left       = 0; 
    capture.crop.c.top        = 0; 
    capture.crop.c.height     = height*24;
    capture.crop.c.width      = width*24;

    /* set the crop area, but don't exit if the device don't support croping */
    xioctl (device, VIDIOC_S_CROP, &capture.crop);

    CLEAR (capture.form);
    capture.form.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    /* read the current setting */
    xioctl (device, VIDIOC_G_FMT, &capture.form);

    /* set the values we want to change */
    capture.form.fmt.pix.width = width; 
    capture.form.fmt.pix.height = height;
    capture.form.fmt.win.chromakey = 0;
    capture.form.fmt.win.field = V4L2_FIELD_ANY;
    capture.form.fmt.win.clips = 0;
    capture.form.fmt.win.clipcount = 0;
    capture.form.fmt.pix.field = V4L2_FIELD_ANY;
    capture.form.fmt.pix.pixelformat = colorspace;

    /* ask the device to change the size*/
    if (-1 == xioctl (device, VIDIOC_S_FMT, &capture.form)) {
      Thread::Err("Could not set specifics of capture window.\n");
      return -1;
    }

    /* Get window info again, to get the real value */
    if (-1 == xioctl (device, VIDIOC_G_FMT, &capture.form)) {
      Thread::Err("Could not obtain specifics of capture window.\n");
      return -1;
    }
    return 0;

}

/*
V4LCamera::V4LCamera(string name,
                     uint8_t camera_index, uint16_t width, uint16_t height,
                     Image::Type::Format format, uint8_t priority)
    : Thread(getFrameworkManager(), name, priority),
      Camera(name, width, height, format) {
  
    string deviceName="/dev/video"+std::to_string(camera_index);
    device = open(deviceName.c_str(), O_RDWR | O_NONBLOCK);
    if (device == -1) {
        Thread::Err("Cannot open %s\n",deviceName.c_str());
    } else {
        Printf("V4LCamera %s, opened %s\n",name.c_str(),deviceName.c_str());
    }

    struct v4l2_capability cap;
    memset(&cap, 0, sizeof (v4l2_capability));
    if (xioctl (device, VIDIOC_QUERYCAP, &cap)==-1) {
        Thread::Err("VIDIOC_QUERYCAP xioctl\n");
    }
    if ((cap.capabilities & V4L2_CAP_VIDEO_CAPTURE) == 0) {
        Thread::Err("device is unable to capture video memory.\n");
    }

    //get v4l2_format
    struct v4l2_format form;
    form.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if(xioctl (device, VIDIOC_G_FMT,&form)==-1) {
        Thread::Err("VIDIOC_G_FMT xioctl\n");
    }
  
    //set width, height and format
    if (format == Image::Type::Format::UYVY) {
        form.fmt.pix.pixelformat = V4L2_PIX_FMT_UYVY; 
    } else if (format == Image::Type::Format::YUYV) {
        form.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV; 
    } else {
        Thread::Err("format not supported\n");
    }
  
    form.fmt.pix.width = width; 
    form.fmt.pix.height = height;
    form.fmt.win.chromakey = 0;
    form.fmt.win.field = V4L2_FIELD_ANY;
    form.fmt.win.clips = 0;
    form.fmt.win.clipcount = 0;
    form.fmt.pix.field = V4L2_FIELD_ANY;
    if(xioctl (device, VIDIOC_S_FMT,&form)==-1) {
        Thread::Err("VIDIOC_S_FMT xioctl\n");
    }
  
    //alloc and queue bufs
    AllocBuffers();
    for (int bufferIndex = 0; bufferIndex < nbBuffers;++bufferIndex) {
        QueueBuffer(bufferIndex);
    }
    
    // enable the streaming
    v4l2_buf_type type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (xioctl (device, VIDIOC_STREAMON,&type)==-1) {
        Thread::Err("VIDIOC_STREAMON xioctl\n");
    }

    // skip first frame. it is often bad -- this is unnotied in traditional apps,
    //  but could be fatal if bad jpeg is enabled
    bufferIndex=-1;
    GrabFrame();

    // ground station
    gain = new DoubleSpinBox(GetGroupBox()->NewRow(), "gain:", 0, 1, 0.1);
    exposure = new DoubleSpinBox(GetGroupBox()->LastRowLastCol(), "exposure:", 0,1, 0.1);
    bright = new DoubleSpinBox(GetGroupBox()->LastRowLastCol(), "bright:", 0, 1, 0.1);
    contrast = new DoubleSpinBox(GetGroupBox()->LastRowLastCol(), "contrast:", 0,1, 0.1);
    hue = new DoubleSpinBox(GetGroupBox()->LastRowLastCol(), "hue:", 0, 1, 0.1);
    sharpness = new DoubleSpinBox(GetGroupBox()->LastRowLastCol(), "sharpness:", 0, 1, 0.1);
    sat = new DoubleSpinBox(GetGroupBox()->LastRowLastCol(), "saturation:", 0, 1,0.1);
    autogain = new CheckBox(GetGroupBox()->NewRow(), "autogain:");
    autoexposure = new CheckBox(GetGroupBox()->LastRowLastCol(), "autoexposure:");
    awb = new CheckBox(GetGroupBox()->LastRowLastCol(), "awb:");
    fps = new Label(GetGroupBox()->NewRow(), "fps");

    hasProblems=false;
}

V4LCamera::~V4LCamera() {
    for (int n_buffers = 0; n_buffers < nbBuffers; n_buffers++) {
       FreeFunction((char*)buffers[n_buffers].start);
   }
    SafeStop();
    Join();
}

void V4LCamera::Run(void) {
  Time cam_time, new_time, fpsNow, fpsPrev;
  int fpsCounter = 0;

  // init image old
  GrabFrame();
  cam_time = GetTime();
  fpsPrev = cam_time;

  while (!ToBeStopped()) {
    //check for ps3eye deconnection in hds uav
    if(hasProblems==false) {
      struct v4l2_format form;
      form.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      xioctl(device, VIDIOC_G_FMT,&form);
      if(xioctl (device, VIDIOC_G_FMT,&form)<0) {
        Thread::Warn("camera disconnected\n");
        hasProblems=true;
      }
    }
    
    // fps counter
    fpsCounter++;
    if (GetTime() > (fpsPrev + 5 * (Time)1000000000)) {
      // every 5 secondes
      fpsNow = GetTime();
      fps->SetText("fps: %.1f",
      fpsCounter / ((float)(fpsNow - fpsPrev) / 1000000000.));
      fpsCounter = 0;
      fpsPrev = fpsNow;
    } 
    
    // cam properties
    if (gain->ValueChanged() == true && autogain->Value() == false)
      SetGain(gain->Value());
    if (exposure->ValueChanged() == true && autoexposure->Value() == false)
      SetExposure(exposure->Value());
    if (bright->ValueChanged() == true)
      SetBrightness(bright->Value());
    if (sat->ValueChanged() == true)
      SetSaturation(sat->Value());
    if (contrast->ValueChanged() == true)
      SetContrast(contrast->Value());
    if (hue->ValueChanged() == true)
      SetHue(hue->Value());
    if (sharpness->ValueChanged() == true)
      SetProperty(V4L2_CID_SHARPNESS, sharpness->Value());
    if (autogain->ValueChanged() == true) {
      if (autogain->Value() == true) {
        gain->setEnabled(false);
      } else {
        gain->setEnabled(true);
        SetGain(gain->Value());
      }
      SetAutoGain(autogain->Value());
    }
    if (autoexposure->ValueChanged() == true) {
      if (autoexposure->Value() == true) {
        exposure->setEnabled(false);
      } else {
        exposure->setEnabled(true);
        SetExposure(exposure->Value());
      }
      SetAutoExposure(autoexposure->Value());
    }
    if (awb->ValueChanged() == true)
      SetProperty(V4L2_CID_AUTO_WHITE_BALANCE, awb->Value());

    // get picture
    GrabFrame();
    new_time = GetTime();
    
    //check for ps3eye deconnection in hds uav
    if(new_time-cam_time>100*1000*1000) {
      Thread::Warn("delta trop grand\n");
      hasProblems=true;
    }

    output->GetMutex();
    output->buffer=(char*)buffers[bufferIndex].start;
    output->ReleaseMutex();

    output->SetDataTime(cam_time);
    ProcessUpdate(output);

    cam_time = new_time;
  }

  close(device);
}
*/
int V4LCamera::allocBuffers() {
   CLEAR (capture.req);
   
   unsigned int buffer_number = DEFAULT_V4L_BUFFERS;

   try_again:
   
   capture.req.count = buffer_number;
   capture.req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
   capture.req.memory = V4L2_MEMORY_MMAP;

   if (-1 == xioctl (device, VIDIOC_REQBUFS, &capture.req)) {
       if (EINVAL == errno) {
         Thread::Warn("does not support memory mapping\n");
       } else {
         perror ("VIDIOC_REQBUFS");
       }
       /* free capture, and returns an error code */
       return -1;
   }

   if (capture.req.count < buffer_number) {
       if (buffer_number == 1) {
           Thread::Warn("Insufficient buffer memory\n");
           /* free capture, and returns an error code */
           return -1;
       } else {
          buffer_number--;
          Thread::Warn("Insufficient buffer memory -- decreaseing buffers\n");
          goto try_again;
       }
   }

   for (n_buffers = 0; n_buffers < capture.req.count; ++n_buffers) {
       struct v4l2_buffer buf;
       CLEAR (buf);

       buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
       buf.memory = V4L2_MEMORY_MMAP;
       buf.index = n_buffers;

       if (-1 == xioctl (device, VIDIOC_QUERYBUF, &buf)) {
           Thread::Warn ("VIDIOC_QUERYBUF error\n");
           /* free capture, and returns an error code */
           return -1;
       }

       capture.buffers[n_buffers].length = buf.length;
       capture.buffers[n_buffers].start =
         mmap (NULL /* start anywhere */,
               buf.length,
               PROT_READ | PROT_WRITE /* required */,
               MAP_SHARED /* recommended */,
               device, buf.m.offset);

       if (MAP_FAILED == capture.buffers[n_buffers].start) {
           Thread::Warn("mmap error\n");
           /* free capture, and returns an error code */
           return -1;
       }
   }

   /* Set up Image data */
   /*
   cvInitImageHeader( &capture.frame,
                      cvSize( capture.captureWindow.width,
                              capture.captureWindow.height ),
                      IPL_DEPTH_8U, 3, IPL_ORIGIN_TL, 4 );*/
   /* Allocate space for RGBA data */
   capture.imageSize=capture.form.fmt.pix.width*capture.form.fmt.pix.height*2;
   capture.imageData = AllocFunction(capture.imageSize);
   Printf("cmem allocated %i at %x\n",capture.imageSize,capture.imageData);
   
   return 1;
};


int V4LCamera::cvGrabFrame(void) {
    unsigned int count;

    count = 1;

    while (count-- > 0) {
        for (;;) {
            fd_set fds;
            struct timeval tv;
            int r;

            FD_ZERO (&fds);
            FD_SET (device, &fds);

            /* Timeout. */
            tv.tv_sec = 2;
            tv.tv_usec = 0;

            r = select (device+1, &fds, NULL, NULL, &tv);

            if (-1 == r) {
                if (EINTR == errno)
                    continue;

                perror ("select");
            }

            if (0 == r) {
                fprintf (stderr, "select timeout\n");

                /* end the infinite loop */
                break;
            }

            if (read_frame_v4l2 (&capture))
                break;
        }
    }
    return(1);
}

int V4LCamera::read_frame_v4l2(CvCaptureCAM_V4L* capture) {
    struct v4l2_buffer buf;

    CLEAR (buf);

    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;

    if (-1 == xioctl (device, VIDIOC_DQBUF, &buf)) {
        switch (errno) {
        case EAGAIN:
            return 0;

        case EIO:
            /* Could ignore EIO, see spec. */

            /* fall through */

        default:
            /* display the error and stop processing */
            perror ("VIDIOC_DQBUF");
            return 1;
        }
   }

   if(buf.index >= capture->req.count) {
     Thread::Err("buf.index >= capture->req.count\n");
   }
   

   capture->bufferIndex = buf.index;


   if (-1 == xioctl (device, VIDIOC_QBUF, &buf))
       perror ("VIDIOC_QBUF");

   return 1;
}

void V4LCamera::Run(void) {
  Time cam_time, new_time, fpsNow, fpsPrev;
  //IplImage *img; // raw image
  int fpsCounter = 0;

  // init image old
  if (!cvGrabFrame()) {
    Printf("Could not grab a frame\n");
  }
  cam_time = GetTime();
  fpsPrev = cam_time;

  while (!ToBeStopped()) {
    // fps counter
    fpsCounter++;
    if (fpsCounter == 100) {
      fpsNow = GetTime();
      fps->SetText("fps: %.1f",
       fpsCounter / ((float)(fpsNow - fpsPrev) / 1000000000.));
      fpsCounter = 0;
      fpsPrev = fpsNow;
    }

    // cam properties
    if (gain->ValueChanged() == true && autogain->Value() == false)
      SetGain(gain->Value());
    if (exposure->ValueChanged() == true && autoexposure->Value() == false)
      SetExposure(exposure->Value());
    if (bright->ValueChanged() == true)
      SetBrightness(bright->Value());
    if (sat->ValueChanged() == true)
      SetSaturation(sat->Value());
    if (contrast->ValueChanged() == true)
      SetContrast(contrast->Value());
    if (hue->ValueChanged() == true)
      SetHue(hue->Value());
    //if (sharpness->ValueChanged() == true)
    //  cvSetCaptureProperty(capture, CV_CAP_PROP_SHARPNESS, sharpness->Value());
    if (autogain->ValueChanged() == true) {
      if (autogain->Value() == true) {
        gain->setEnabled(false);
      } else {
        gain->setEnabled(true);
        SetGain(gain->Value());
      }
      SetAutoGain(autogain->Value());
    }
    if (autoexposure->ValueChanged() == true) {
      if (autoexposure->Value() == true) {
        exposure->setEnabled(false);
      } else {
        exposure->setEnabled(true);
        SetExposure(exposure->Value());
      }
      SetAutoExposure(autoexposure->Value());
    }
    //if (awb->ValueChanged() == true)
    //  cvSetCaptureProperty(capture, CV_CAP_PROP_AWB, awb->Value());
   
    // cam pictures
    cvRetrieveRawFrame(&capture);
    if (!cvGrabFrame()) {
      Printf("Could not grab a frame\n");
    }
    new_time = GetTime();
        
    //check for ps3eye deconnection in hds uav
    if(new_time-cam_time>100*1000*1000) {
      Thread::Warn("delta trop grand\n");
      hasProblems=true;
    }

    output->GetMutex();
    
    output->buffer=capture.imageData;
    output->ReleaseMutex();
   
    output->SetDataTime(cam_time);
    ProcessUpdate(output);
    cam_time = new_time;
  }

  //cvReleaseCapture(&capture);
}

void V4LCamera::cvRetrieveRawFrame( CvCaptureCAM_V4L* capture) {




   /* Now get what has already been captured as a IplImage return */

   /* First, reallocate imageData if the frame size changed */


/*
    if(((unsigned long)capture->frame.width != capture->form.fmt.pix.width)
       || ((unsigned long)capture->frame.height != capture->form.fmt.pix.height)) {
           if (PALETTE_YUYV == 1 || PALETTE_UYVY == 1)
           {
                cvFree(&capture->frame.imageData);
                cvInitImageHeader( &capture->frame,
                cvSize( capture->form.fmt.pix.width,
                capture->form.fmt.pix.height ),
                IPL_DEPTH_8U,2, IPL_ORIGIN_TL, 4 );
                capture->frame.imageData = (char *)cvAlloc(capture->frame.imageSize);
           }else
           {
              fprintf( stderr,
                 "HIGHGUI ERROR: V4L: raw output not supported for this palette\n");
           }

    }

 */


    
  memcpy((char *)capture->imageData,(char *)capture->buffers[capture->bufferIndex].start,capture->imageSize);
    
 

   
}
int V4LCamera::QueueBuffer(int index) {
    struct v4l2_buffer buf;
    if(index>=0 && index<nbBuffers) {
        memset(&buf, 0, sizeof (v4l2_buffer));
        buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory      = V4L2_MEMORY_USERPTR;//V4L2_MEMORY_MMAP;
        buf.index       = (unsigned long)index;
        buf.m.userptr=(unsigned long)(buffers[index].start);
        buf.length=buffers[index].length;
        
        int ret=xioctl (device, VIDIOC_QBUF, &buf);
        if (ret==-1) {
            Thread::Err("VIDIOC_QBUF xioctl %s\n",strerror(-ret));
            return -1;
        }
    }
    return 0;
}

int V4LCamera::GrabFrame(void) {
    //queue previous buffer
    if(QueueBuffer(bufferIndex)<0) return -1;
   
    fd_set fds;
    struct timeval tv;
    FD_ZERO (&fds);
    FD_SET (device, &fds);

    tv.tv_sec = 0;
    tv.tv_usec = 100000;

    int r = select (device+1, &fds, NULL, NULL, &tv);

    if (-1 == r) {
        char errorMsg[256];
        Thread::Err("select (%s)\n", strerror_r(-r, errorMsg, sizeof(errorMsg)));
        return -1;
    }

    if (0 == r) {
        Thread::Err("select timeout\n");
        return -1;
    }

    struct v4l2_buffer buf;
    memset(&buf, 0, sizeof (v4l2_buffer));
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_USERPTR;//V4L2_MEMORY_MMAP;

    //get last captured image
    int prevDQbuf=-1;
    for(int i=0;i<4;i++) {
        if (xioctl (device, VIDIOC_DQBUF, &buf)==-1) {
            if (errno==EAGAIN) {
                break;
            } else {
                Thread::Err("VIDIOC_DQBUF xioctl\n");
                return -1;
            }
       } else {
           if(prevDQbuf!=-1) {
               QueueBuffer(prevDQbuf);
           }
           for (int n_buffers = 0; n_buffers < nbBuffers; n_buffers++) {
               if((void*)(buf.m.userptr)==buffers[n_buffers].start) {
                   prevDQbuf=n_buffers;
                   bufferIndex=n_buffers;
                   break;
               }
           }
       }
    }
   
   return 1;
}

int V4LCamera::AllocBuffers(void) {
    struct v4l2_requestbuffers requestbuffers;
   memset(&requestbuffers, 0, sizeof (v4l2_requestbuffers));
   
   unsigned int buffer_number = DEFAULT_V4L_BUFFERS;

   try_again:
   
   requestbuffers.count = buffer_number;
   requestbuffers.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
   requestbuffers.memory = V4L2_MEMORY_USERPTR;//V4L2_MEMORY_MMAP;

   if (xioctl (device, VIDIOC_REQBUFS, &requestbuffers)==-1) {
       if (errno==EINVAL) {
         Thread::Err("VIDIOC_REQBUFS user memory not supported\n");
       } else {
         Thread::Err ("VIDIOC_REQBUFS xioctl\n");
       }
       return -1;
   }
   
   nbBuffers=DEFAULT_V4L_BUFFERS;
   for (int n_buffers = 0; n_buffers < nbBuffers; n_buffers++) {
       buffers[n_buffers].length = output->GetDataType().GetSize();
       buffers[n_buffers].start =AllocFunction(output->GetDataType().GetSize());
   }

   return 1;
};

bool V4LCamera::HasProblems(void) {
  return hasProblems;
}

void V4LCamera::SetAutoGain(bool value) {
  SetProperty(V4L2_CID_AUTOGAIN, value);
}

void V4LCamera::SetAutoExposure(bool value) {
  Thread::Warn("not implemented\n");
}

void V4LCamera::SetGain(float value) {
  SetProperty(V4L2_CID_GAIN, value);
}

void V4LCamera::SetExposure(float value) {
  SetProperty(V4L2_CID_EXPOSURE, value);
}

void V4LCamera::SetBrightness(float value) {
  SetProperty(V4L2_CID_BRIGHTNESS, value);
}

void V4LCamera::SetSaturation(float value) {
  SetProperty(V4L2_CID_SATURATION, value);
}

void V4LCamera::SetHue(float value) {
  SetProperty(V4L2_CID_HUE, value);
}

void V4LCamera::SetContrast(float value) {
  SetProperty(V4L2_CID_CONTRAST, value);
}

float V4LCamera::GetProperty(int property) {
  //get min and max value
  struct v4l2_queryctrl queryctrl;
  queryctrl.id = property;
  if(xioctl (device, VIDIOC_QUERYCTRL,&queryctrl)==-1) return -1;
  int min = queryctrl.minimum;
  int max = queryctrl.maximum;
 
  //set value
  struct v4l2_control control;
  memset (&control, 0, sizeof (v4l2_control));
  control.id = property;
  if(xioctl (device,VIDIOC_G_CTRL, &control)==-1) return -1;
  
  return ((float)control.value - min + 1) / (max - min);
}

void V4LCamera::SetProperty(int property,float value) {
  //get min and max value
  struct v4l2_queryctrl queryctrl;
  queryctrl.id = property;
  if(xioctl (device, VIDIOC_QUERYCTRL,&queryctrl)==-1) {
    Thread::Warn("prop %x, VIDIOC_QUERYCTRL failed\n",property);
  }
  int min = queryctrl.minimum;
  int max = queryctrl.maximum;
 
  //set value
  struct v4l2_control control;
  memset (&control, 0, sizeof (v4l2_control));
  control.id = property;
  control.value = (int)(value * (max - min) + min);
  if(xioctl (device,VIDIOC_S_CTRL, &control)==-1) {
    Thread::Warn("prop %x, VIDIOC_S_CTRL failed\n",property);
  }
}

int V4LCamera::xioctl( int fd, int request, void *arg) {
  int r;

  do r = ioctl (fd, request, arg);
  while (-1 == r && EINTR == errno);

  return r;
}

} // end namespace sensor
} // end namespace flair
